//
//   File : kvi_expreval.cpp (/usr/build/NEW_kvirc/kvirc/src/kvilib/kvi_expreval.cpp)
//   Last major modification : Sun Feb 21 1999 15:05:41 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1998-1999 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
#include "kvi_expreval.h"

//
// A (working!) class to parse algebraic expression strings.
// Excluded just because I've integrated it in the KviUserParser
// and expanded it with string comparisons.
// Uses the KviStr class from libkvilib , but can be easily
// hacked to use another string class.
//
// Usage:
// const char * expression = "2 * 7 - 3 < 4 && 7 + 9 | ~4 - -7 / 9 || 4 +(7 - 3 <= (3*6 != 1)+ 5*1 == 3)";
// KviAlgExpr e(expression);
// int ret = e->evaluate(returnBufferString);
// // or overload : int ret = e->evaluate(&returnLong);
// if (ret != KVI_EXPR_OK){
//    check tre error code...
// } else {
//    // returnLong contains the result of calculations
//    // or return bufferString contains the same result
//    // but in form of....yes , string.
// }
//

// gnary hack for pow() 
// (just to avoid to include math.h and have to handle the range errors)
long my_funky_pow(long x,long y)
{
	long total = x;
	for(int i=1;i<y;i++)total = total*x;
	return total;
}

//===========================================================================//
// Expression tree node class
// m_type describes the node type : operand, binary or unary operator
// m_left and m_right are children of this node
// m_value is the value of the operand or type of operator.
// m_parent is the parent node , or 0 if this is a toplevel node
//

KviExprTreeNode::KviExprTreeNode(NodeType t,long v)
{
	m_type = t;
	m_left = 0;
	m_right = 0;
	m_value = v;
	m_parent = 0;
}


KviExprTreeNode::~KviExprTreeNode()
{
	if(m_left)delete m_left;
	if(m_right)delete m_right;
}

void KviExprTreeNode::setLeft(KviExprTreeNode * node)
{
	m_left = node;
	node->m_parent = this;
}

void KviExprTreeNode::setRight(KviExprTreeNode * node)
{
	m_right = node;
	node->m_parent = this;
}

long KviExprTreeNode::calculate()
{
	if(m_type == Operand)return m_value;
	if(m_type == UnaryOperator){
		switch(m_value){
			case OP_NOT:    return ! m_left->calculate(); break;
			case OP_NEGATE: return - m_left->calculate(); break;
			case OP_BITNOT: return ~ m_left->calculate(); break;
		}
	}
	switch(m_value){
		case OP_BITAND:
			return (m_left->calculate() & m_right->calculate());
		break;
		case OP_BITOR:
			return (m_left->calculate() | m_right->calculate());
		break;
		case OP_SHR:
			return (m_left->calculate() >> m_right->calculate());
		break;
		case OP_SHL:
			return (m_left->calculate() << m_right->calculate());
		break;
		case OP_POWER:
			return (long)my_funky_pow(m_left->calculate(),m_right->calculate());
		break;
		case OP_MUL:
			return (m_left->calculate() * m_right->calculate());
		break;
		case OP_DIV: {
			long temp = m_right->calculate();
			if(temp == 0)return 0; //Division by zero
			return (m_left->calculate() / temp);
		}
		break;
		case OP_MOD:{
			long temp = m_right->calculate();
			if(temp == 0)return 0;
			return (m_left->calculate() % temp);
		}
		break;
		case OP_ADD:
			return (m_left->calculate() + m_right->calculate());
		break;
		case OP_SUB:
			return (m_left->calculate() - m_right->calculate());
		break;
		case OP_GT:
			return (m_left->calculate() > m_right->calculate());
		break;
		case OP_LT:
			return (m_left->calculate() < m_right->calculate());
		break;
		case OP_EQ:
			return (m_left->calculate() == m_right->calculate());
		break;
		case OP_GE:
			return (m_left->calculate() >= m_right->calculate());
		break;
		case OP_LE:
			return (m_left->calculate() <= m_right->calculate());
		break;
		case OP_NE:
			return (m_left->calculate() != m_right->calculate());
		break;
		case OP_AND:
			return (m_left->calculate() && m_right->calculate());
		break;
		case OP_OR:
			return (m_left->calculate() || m_right->calculate());
		break;
	}
}

//void KviExprTreeNode::dump()
//{
//	if(m_left)m_left->dump();
//
//	switch(m_type){
//		case BinaryOperator:
//			debug("Node : Binary operator %d",m_value);
//		break;
//		case UnaryOperator:
//			debug("Node : Unary operator %d",m_value);
//		break;
//		default:
//			debug("Node : Operand %d",m_value);	
//		break;
//	}
//
//	if(m_right)m_right->dump();
//}

KviExprTreeNode * KviExprTreeNode::parentWithPrecedenceLowerThan(long operValue)
{
	if(m_value >= operValue)return this;
	if(m_parent == 0)return 0;
	return m_parent->parentWithPrecedenceLowerThan(operValue);
}

long KviExprTreeNode::firstBinaryOperator()
{
	if(m_type == BinaryOperator)return m_value;
	return m_left->firstBinaryOperator();
}

//===========================================================================//
// Expression parser
// each instande of the class can be used only once!...
// (easy to modify...need just some cleanup when exiting from evaluate())
//
//

KviAlgExpr::KviAlgExpr(const char *pData)
{
	m_root = 0;
	m_ptr = pData;
	m_error = KVI_EXPR_OK;
}

KviAlgExpr::~KviAlgExpr()
{
	if(m_root)delete m_root;
}

int KviAlgExpr::evaluate(KviStr &buffer)
{
	m_root = evaluateExpression();
	if(!m_root)return m_error;

	if(*m_ptr != '\0')return KVI_EXPR_PARENTHESIS_MISMATCH;

	buffer.setNum(m_root->calculate());
	return m_error;
}

int KviAlgExpr::evaluate(long *valuePtr)
{
	m_root = evaluateExpression();
	if(!m_root)return m_error;

	if(*m_ptr != '\0')return KVI_EXPR_PARENTHESIS_MISMATCH;

	*valuePtr = m_root->calculate();
	return m_error;
}

KviExprTreeNode * KviAlgExpr::evaluateExpression()
{
	// This one evaluates the expression up to a (not matching) ')' or a \0
	KviExprTreeNode * op = evaluateOperand();
	if(!op)return 0;

	skipSpaces();
	if(*m_ptr == ')' || *m_ptr == '\0')return op; //single operand

	KviExprTreeNode * curOper = evaluateOperator();
	if(!curOper){
		delete op;
		return 0;
	}

	curOper->setLeft(op);

	return evaluateExpression_RightOperand(curOper);
}

KviExprTreeNode * KviAlgExpr::evaluateExpression_RightOperand(KviExprTreeNode * curTopOperator)
{
	// Now curTopOperator has the left subtree (one node) set
	// and it points to the TOP (=ROOT) node
	// Evaluate the rest
	KviExprTreeNode * operand;
	KviExprTreeNode * incompleteOperator = curTopOperator;
	KviExprTreeNode * auxOperator;

	for(;;){
		operand = evaluateOperand();
		if(!operand){ delete curTopOperator; return 0; }
		skipSpaces();
		if(*m_ptr == ')' || *m_ptr == '\0'){ incompleteOperator->setRight(operand); return curTopOperator; }

		auxOperator = evaluateOperator();
		if(!auxOperator){ delete curTopOperator; delete operand; return 0; }

		//now compare operators...
		if(incompleteOperator->firstBinaryOperator() >= auxOperator->firstBinaryOperator()){
			incompleteOperator->setRight(auxOperator);
			auxOperator->setLeft(operand);
		} else {
			// precedence of operators...
			incompleteOperator->setRight(operand); //right tree complete
			// go up until we find an operator with lower precedence than auxOperator (>=)
			KviExprTreeNode * tempOperator = incompleteOperator->parentWithPrecedenceLowerThan(auxOperator->m_value);
			if(tempOperator == 0){ //reached the top
				auxOperator->setLeft(curTopOperator);
				curTopOperator = auxOperator;
			} else {
				auxOperator->setLeft(tempOperator->right());
				tempOperator->setRight(auxOperator);
			}
		}
		incompleteOperator = auxOperator;
	}
}

KviExprTreeNode * KviAlgExpr::evaluateOperator()
{
	// This one extracts the first operator it finds	
	skipSpaces();
	long op;
	switch(*m_ptr){
		// single ops
		case '^': op = OP_POWER; m_ptr++; break;
		case '*': op = OP_MUL;   m_ptr++; break;
		case '/': op = OP_DIV;   m_ptr++; break;
		case '%': op = OP_MOD;   m_ptr++; break;
		case '+': op = OP_ADD;   m_ptr++; break;
		case '-': op = OP_SUB;   m_ptr++; break;
		case '=':
			m_ptr++;
			if(*m_ptr == '=')op = OP_EQ , m_ptr++;
			else {
				m_error = KVI_EXPR_UNKNOWN_OPERATOR;
				return 0;
			}
		break;
		case '!':
			m_ptr++;
			if(*m_ptr == '=')op = OP_NE , m_ptr++;
			else {
				m_error = KVI_EXPR_UNKNOWN_OPERATOR;
				return 0;
			}
		break;
		case '>':
			m_ptr++;
			if(*m_ptr == '=')op = OP_GE , m_ptr++;
			else if(*m_ptr == '>')op = OP_SHR , m_ptr++;
			else op = OP_GT;
		break;
		case '<':
			m_ptr++;
			if(*m_ptr == '=')op = OP_LE , m_ptr++;
			else if(*m_ptr == '<')op = OP_SHL , m_ptr++;
			else op = OP_LT;
		break;
		case '&':
			m_ptr++;
			if(*m_ptr == '&')op = OP_AND , m_ptr++;
			else op = OP_BITAND;
		break;
		case '|':
			m_ptr++;
			if(*m_ptr == '|')op = OP_OR  , m_ptr++;
			else op = OP_BITOR;
		break;
		default:
			m_error = KVI_EXPR_UNKNOWN_OPERATOR;
			return 0;
		break;
	}

	return new KviExprTreeNode(KviExprTreeNode::BinaryOperator,op);
}


KviExprTreeNode * KviAlgExpr::evaluateOperand()
{
	// This one tries to extract a numeric operand from the data buffer
	// Moves m_ptr to the character after the last char of the operand
	// Returns a valid KviExprTreeNode of succesful 
	// otherwise returns 0 and sets m_error appropriately

	skipSpaces();

	KviExprTreeNode * pTop = 0;
	KviExprTreeNode * pCur = 0;
	KviExprTreeNode * pLast = 0;
	// Check for unary ops...

	while(*m_ptr == '-' || *m_ptr == '~' || *m_ptr == '!'){
		// Unary op
		switch(*m_ptr){
			case '-':
				pCur = new KviExprTreeNode(KviExprTreeNode::UnaryOperator,OP_NEGATE);
			break;
			case '~':
				pCur = new KviExprTreeNode(KviExprTreeNode::UnaryOperator,OP_BITNOT);
			break;
			case '!':
				pCur = new KviExprTreeNode(KviExprTreeNode::UnaryOperator,OP_NOT);
			break;
		}
		// And we work with the child
		if(pTop == 0)pTop = pCur;
		else pLast->setLeft(pCur);
		pLast = pCur;
		m_ptr++;
		skipSpaces();
	}

	if(*m_ptr == '('){
		// subexpression
		m_ptr++;
		pCur = evaluateExpression();
		if(!pCur){                              // failed somewhere in the call
			if(pTop)delete pTop;
			return 0;
		}
		if(*m_ptr != ')'){                         // failed here
			if(pTop)delete pTop;
			delete pCur;
			m_error = KVI_EXPR_MISSING_CLOSING_PARENTHESIS;
			return 0;
		}
		m_ptr++;
	} else {
		// here *m_ptr may be something that is not a digit too!
		pCur = evaluateSimpleOperand();
		if(!pCur){                            // failed somewere
			if(pTop)delete pTop;
			return 0;
		}
	}
	if(pTop)pLast->setLeft(pCur);
	else return pCur;
	return pTop;
}

KviExprTreeNode * KviAlgExpr::evaluateSimpleOperand()
{
	// This one extracts a number from the data buffer
	if(!isxdigit(*m_ptr)){
		m_error = KVI_EXPR_UNEXPECTED_CHAR;
		return 0;
	}
	int base = 10;
	if(*m_ptr == '0'){
		m_ptr++;
		switch(tolower(*m_ptr)){
			case 'x': m_ptr++; base = 16; break;
			case 'h': m_ptr++; base = 16; break;

			case 'i': m_ptr++; base = 2;  break; // can not use b because it is a hex digit
			case 'j': m_ptr++; base = 3;  break;
			case 'k': m_ptr++; base = 4;  break;
			case 'l': m_ptr++; base = 5;  break;
			case 'm': m_ptr++; base = 6;  break;
			case 'n': m_ptr++; base = 7;  break;
			case 'o': m_ptr++; base = 8;  break;
			case 'p': m_ptr++; base = 9;  break;
			case 'q': m_ptr++; base = 10; break;
			case 'r': m_ptr++; base = 11; break;
			case 's': m_ptr++; base = 12; break;
			case 't': m_ptr++; base = 13; break;
			case 'u': m_ptr++; base = 14; break;
			case 'v': m_ptr++; base = 15; break;
			default:  m_ptr--; break; // unknown ...maybe a single 0
		}
		if((!isxdigit(*m_ptr))){
			m_error = KVI_EXPR_UNEXPECTED_CHAR;
			return 0;
		}
	}
	const char *aux = m_ptr;
	while(isxdigit(*m_ptr))m_ptr++;

	KviStr num(aux,m_ptr - aux);
	return new KviExprTreeNode(KviExprTreeNode::Operand,num.toLongExt(0,base));
}

