//
//   File : kvidns.cpp (/usr/build/NEW_kvirc/kvirc/kvidns/kvidns.cpp)
//   Last major modification : Sun Jan 10 1999 15:21:04 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1998-1999 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

//
//   DNS module for KVirc
//
//   Syntax :
//            kvidns <query_hostname_or_ip>
//   Output :
//            ERR:<query>:<error_character_defined_in_kvidns_h>\n
//            DNS:<query>:<hostname>:<ip>:<alias1 | *>:<alias2 | *>\n
//            All output to stdout.
//

#include "../common/kvi_settings.h"

#include "kvidns.h"

#include <stdarg.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>

//Check for inet aton

#ifndef HAVE_INET_ATON

#warning "Your system lacks the inet_aton function,"
#warning "or you're trying to compile this file without"
#warning "the config.h created by the configure script."
#warning "Using own internal implementation of inet_aton."

#include <ctype.h>

// Need own inet_aton implementation

// 
// Check whether "cp" is a valid ascii representation
// of an Internet address and convert to a binary address.
// Returns 1 if the address is valid, 0 if not.
// This replaces inet_addr, the return value from which
// cannot distinguish between failure and a local broadcast address.
//
// From ircd source.
//

int my_inet_aton(register const char *cp, struct in_addr *addr)
{
	register unsigned long val;
	register int base, n;
	register char c;
	unsigned int parts[4];
	register unsigned int *pp = parts;

	c = *cp;
	for(;;){
		//
		// Collect number up to ``.''.
		// Values are specified as for C:
		// 0x=hex, 0=octal, isdigit=decimal.
		//
		if(!isdigit(c))return (0);
		val = 0; base = 10;
		if (c == '0') {
			c = *++cp;
			if (c == 'x' || c == 'X')base = 16, c = *++cp;
			else base = 8;
		}
		for (;;) {
			if (isascii(c) && isdigit(c)) {
				val = (val * base) + (c - '0');
				c = *++cp;
			} else if (base == 16 && isascii(c) && isxdigit(c)) {
				val = (val << 4) | (c + 10 - (islower(c) ? 'a' : 'A'));
				c = *++cp;
			} else break;
		}
		if (c == '.') {
			//
			// Internet format:
			//	a.b.c.d
			//	a.b.c	(with c treated as 16 bits)
			//	a.b	(with b treated as 24 bits)
			//
			if (pp >= parts + 3) return (0);
			*pp++ = val;
			c = *++cp;
		} else break;
	}
	// Check for trailing characters.
	if (c != '\0' && (!isascii(c) || !isspace(c)))return (0);
	//
	// Concact the address according to
	// the number of parts specified.
	//
	n = pp - parts + 1;
	switch (n) {
	case 0:
		return (0);		/* initial nondigit */
	case 1:				/* a -- 32 bits */
		break;
	case 2:				/* a.b -- 8.24 bits */
		if (val > 0xffffff) return (0);
		val |= parts[0] << 24;
		break;
	case 3:				/* a.b.c -- 8.8.16 bits */
		if (val > 0xffff)return (0);
		val |= (parts[0] << 24) | (parts[1] << 16);
		break;
	case 4:				/* a.b.c.d -- 8.8.8.8 bits */
		if (val > 0xff)return (0);
		val |= (parts[0] << 24) | (parts[1] << 16) | (parts[2] << 8);
		break;
	}
	if (addr)addr->s_addr = htonl(val);
	return (1);
}

#else
	#define my_inet_aton(_arg1,_arg2) inet_aton(_arg1,_arg2)
#endif //HAVE_INET_ATON

void writeError(const char *query,char error)
{
	fprintf(stdout,"ERR:%s:%c\n",query,error);
}

int main( int argc , char **argv)
{
	if(argc<2){
		writeError("???",KVIDNS_NO_HOST_TO_RESOLVE);
		return 0;
	}

	char *hostToResolve = argv[1];
	struct in_addr inAddr;
	struct hostent *hostEntry;

	if(my_inet_aton(hostToResolve,&inAddr)){
		hostEntry=gethostbyaddr((const char *)&inAddr,sizeof(inAddr),AF_INET);
	} else {
		hostEntry=gethostbyname(hostToResolve);
	}
	if(!hostEntry){
		switch(h_errno){
			case HOST_NOT_FOUND:
				writeError(hostToResolve,KVIDNS_HOST_NOT_FOUND);
				break;
			case NO_ADDRESS:
				writeError(hostToResolve,KVIDNS_VALID_NAME_BUT_NO_IP);
				break;
			case NO_RECOVERY:
				writeError(hostToResolve,KVIDNS_UNRECOVERABLE_NAMESERVER_ERROR);
				break;
			case TRY_AGAIN:
				writeError(hostToResolve,KVIDNS_TEMPORANEOUS_FAULT);
				break;
			default:
				writeError(hostToResolve,KVIDNS_DNS_FAILURE);
				break;
		}
		return 0;
	}
	if(hostEntry->h_aliases[0]){
		if(hostEntry->h_aliases[1]){
			fprintf(stdout,"DNS:%s:%s:%s:%s:%s\n",
							hostToResolve,
							hostEntry->h_name,
							inet_ntoa( * (struct in_addr*) hostEntry->h_addr),
							hostEntry->h_aliases[0],
							hostEntry->h_aliases[1]);	

		} else {
			fprintf(stdout,"DNS:%s:%s:%s:%s:*\n",
							hostToResolve,
							hostEntry->h_name,
							inet_ntoa( * (struct in_addr*) hostEntry->h_addr),
							hostEntry->h_aliases[0]);	
		}
	} else {
		fprintf(stdout,"DNS:%s:%s:%s:*:*\n",
							hostToResolve,
							hostEntry->h_name,
							inet_ntoa( * (struct in_addr*) hostEntry->h_addr));	
	}
	return 0;
}
