//
//   File : kvi_channel.cpp (/usr/build/NEW_kvirc/kvirc/src/kvirc/kvi_channel.cpp)
//   Last major modification : Sat Apr 3 1999 17:19:37 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#define _KVI_DEBUG_CHECK_RANGE_
#include "kvi_debug.h"

#include "kvi_channel.h"
#include "kvi_app.h"
#include "kvi_ircview.h"
#include "kvi_input.h"
#include "kvi_frame.h"
#include "kvi_listbox.h"
#include "kvi_locale.h"
#include "kvi_options.h"
#include "kvi_uparser.h"

#include "kvi_fileutils.h"

#include "kvi_popupmenu.h"

#include "kvi_ircsocket.h"
#include "kvi_chaneditors.h"

#include "kvi_event.h"

#include <qtooltip.h>
#include <qpalette.h>
#include <qdatetime.h>
#include <qcursor.h>
#include <qvaluelist.h>
#include <qnamespace.h>
#include <qtimer.h>

/*
	@quickhelp: KviChannel
	@widget: Channel window
		This window is the rappresentation of the IRC channel.<br>
		The channel name is displayed in the top label of the window.<br>
		For scripting purposes the channel name is also the window name
		reported by the <a href="s_window.kvihelp">$window</a> identifier.<br>
		Just under the channel name you can see three labels showing (from left to right):
		the <a href="qh_kvichanlabel_topic.kvihelp">channel topic</a>, the <a href="qh_kvichanlabel_mode.kvihelp">current channel mode</a>
		and the <a href="qh_kvichanlabel_users.kvihelp">users statistics</a>.<br>
		The main part of the window is the <a href="qh_kviircview.kvihelp">channel view</a>
		and on the right of that you have the <a href="qh_kvilistbox.kvihelp">channel members list</a>.<br>
		The lowest part of the window is the <a href="qh_kviinput.kvihelp">input edit</a>.<br>
		To leave the channel press the close button in the upper right corner or use the <a href="part.kvihelp">part</a> command.<br>
*/

//Declared in kvi_app.cpp and managed by KviApp class
extern QPixmap * g_pWindowIcon[KVI_WND_NUM_ICONS];
extern KviPopupMenu * g_pChannelPopup;
extern KviPopupMenu * g_pUserlistPopup;

extern KviEventManager * g_pEventManager;

KviChanLabel::KviChanLabel(const QString &text,QWidget *parent,const char *name)
:QLabel(text,parent,name){ setFrameStyle(QFrame::Panel|QFrame::Sunken); m_bCanEdit = false; }
KviChanLabel::~KviChanLabel(){}
void KviChanLabel::mousePressEvent(QMouseEvent *e){
	if(e->button() & RightButton)emit rightClicked();
	else QLabel::mousePressEvent(e);
}
void KviChanLabel::mouseDoubleClickEvent(QMouseEvent *){ m_bCanEdit = true; }
void KviChanLabel::mouseReleaseEvent(QMouseEvent *)
{ if(m_bCanEdit){ emit doubleClicked(); m_bCanEdit = false; } }


//============ KviChannel ============//

KviChannel::KviChannel(KviFrame *lpFrm,const char * name)
:KviWindow(name,KVI_WND_TYPE_CHANNEL,lpFrm)
{
	m_pBanList     = new QList<KviBanMask>;
	m_pBanList->setAutoDelete(true);
	m_iBanCount    = 0;
	m_pExceptionList = new QList<KviBanMask>;
	m_pExceptionList->setAutoDelete(true);
	m_iExceptionCount = 0;
	m_pInviteExceptionList = new QList<KviBanMask>;
	m_pInviteExceptionList->setAutoDelete(true);
	m_iInviteExceptionCount = 0;


	m_pTopSplitter = new QSplitter(QSplitter::Horizontal,this);
	m_pTopicLabel  = new KviChanLabel(__tr("[ Channel topic unknown ]"),m_pTopSplitter,"topic");
	QToolTip::add(m_pTopicLabel,QString(__tr("[ Channel topic unknown ]\n(No TOPIC message was received from server yet)")));
	m_pTopicLabel->setTextFormat(Qt::PlainText);
	connect(m_pTopicLabel,SIGNAL(doubleClicked()),this,SLOT(doEditTopic()));
	m_pModeLabel   = new KviChanLabel("",m_pTopSplitter,"mode");
	m_pModeLabel->setTextFormat(Qt::PlainText);
	connect(m_pModeLabel,SIGNAL(doubleClicked()),this,SLOT(doEditModes()));
	m_pUsersLabel  = new KviChanLabel("0 [o:0 v:0 b:0]",m_pTopSplitter,"users");
	m_pUsersLabel->setFrameStyle(QFrame::Panel|QFrame::Sunken);
	m_pUsersLabel->setTextFormat(Qt::PlainText);
	connect(m_pUsersLabel,SIGNAL(doubleClicked()),this,SLOT(doEditMasks()));
	connect(m_pUsersLabel,SIGNAL(rightClicked()),this,SLOT(usersLabelRightClicked()));
	m_pSplitter    = new QSplitter(QSplitter::Horizontal,this);
	m_pView        = new KviIrcView(m_pSplitter,lpFrm,this);
	m_pListBox     = new KviListBox(m_pSplitter,this,lpFrm->m_pUserParser,lpFrm);
	connect(m_pListBox,SIGNAL(rightClicked()),this,SLOT(listBoxRightClicked()));


	m_pInput       = new KviInput(this,lpFrm->m_pUserParser,m_pView,m_pListBox);
	// reasonable default for the splitters
	QValueList<int> l;
	l.append(80); //%
	l.append(20); //%
	m_pSplitter->setSizes(l);

	QValueList<int> l2;
	l2.append(60); //%
	l2.append(20); //%
	l2.append(20); //%
	m_pTopSplitter->setSizes(l2);

	setFocusHandler(m_pInput);
	applyLabelsOptions();

	connect(m_pView,SIGNAL(contextPopupRequested(KviIrcView *)),this,SLOT(viewRightClicked(KviIrcView *)));
	m_pTopicEditor = 0; //create it later...when requested
	m_pModeEditor  = 0; //same as above
	m_pMaskEditor  = 0;
	m_bTopicChangeEnabled = true;
	m_bModeChangeEnabled  = true;
	m_joinTime     = QTime::currentTime();
	m_iChanSyncMask= KVI_CHANSYNC_NOTINSYNC;
	m_pMasksEditPopup = 0;
	m_iLastMasksEdited = -1;
}

//============ ~KviChannel ============//

KviChannel::~KviChannel()
{
	delete m_pBanList;
	delete m_pExceptionList;
	delete m_pInviteExceptionList;
}

void KviChannel::saveProperties()
{
	KviWindowProperty p;
	p.rect = externalGeometry();
	p.isDocked = isAttacched();
	QValueList<int> l(m_pSplitter->sizes());
	if(l.count() >= 1)p.splitWidth1 = *(l.at(0));
	if(l.count() >= 2)p.splitWidth2 = *(l.at(1));
	QValueList<int> l2(m_pTopSplitter->sizes());
	if(l2.count() >= 1)p.topSplitWidth1 = *(l2.at(0));
	if(l2.count() >= 2)p.topSplitWidth2 = *(l2.at(1));
	if(l2.count() >= 3)p.topSplitWidth3 = *(l2.at(2));
	p.timestamp = m_pView->timestamp();
	p.imagesVisible = m_pView->imagesVisible();
	p.isMaximized = isAttacched() && isMaximized();
	g_pOptions->m_pWinPropertiesList->setProperty(caption(),&p);
}

void KviChannel::setProperties(KviWindowProperty *p)
{
	QValueList<int> l;
	l.append(p->splitWidth1);
	l.append(p->splitWidth2);
	m_pSplitter->setSizes(l);
	QValueList<int> l2;
	l2.append(p->topSplitWidth1);
	l2.append(p->topSplitWidth2);
	l2.append(p->topSplitWidth3);
	m_pTopSplitter->setSizes(l2);
	m_pView->setTimestamp(p->timestamp);
	m_pView->setShowImages(p->imagesVisible);
}

void KviChannel::doEditTopic()
{
	if(!m_bTopicChangeEnabled)return;
	if(!m_pTopicEditor){
		m_pTopicEditor = new KviTopicEditor(this);
//		m_pTopicEditor->hide();
		m_pTopicEditor->connectMe(this,SLOT(topicEditCompleted()));
	}
	QPoint p = mapFromGlobal(m_pTopicLabel->mapToGlobal(QPoint(0,0)));
	m_pTopicEditor->setFont(m_pTopicLabel->font());
	m_pTopicEditor->setGeometry(p.x(),p.y(),m_pTopicLabel->width(),m_pTopicLabel->height());
	m_pTopicEditor->setText(m_pTopicLabel->text());
	m_pTopicEditor->show();
	m_pTopicEditor->setFocus();
}

void KviChannel::topicEditCompleted()
{
	m_pTopicEditor->hide();
	KviStr tmp1 = m_pTopicEditor->text();
	KviStr tmp2 = m_pTopicLabel->text();
	if(!kvi_strEqualCS(tmp1.ptr(),tmp2.ptr())){
		m_pFrm->m_pSocket->sendFmtData("TOPIC %s :%s",caption(),tmp1.ptr());
	}
	m_pInput->setFocus();
}

void KviChannel::doEditModes()
{
	if(!m_bModeChangeEnabled)return;
	if(!m_pModeEditor){
		m_pModeEditor  = new KviModeEditor(this);
		m_pModeEditor->connectMe(this,SLOT(modeEditCompleted(const char *)));
	}
	QPoint p=mapFromGlobal(m_pModeLabel->mapToGlobal(QPoint(0,m_pModeLabel->height())));
	int hght = m_pModeEditor->sizeHint().height();
	if(hght > (height() - m_pModeLabel->height()))hght = height() - m_pModeLabel->height();
	m_pModeEditor->setGeometry(p.x(),p.y(),m_pModeLabel->width(),hght);
	KviStr tmp = m_pModeLabel->text();
	m_pModeEditor->doEdit(tmp);
}

void KviChannel::modeEditCompleted(const char * modeDiff)
{
//	m_pModeEditor->hide(); // not needed
	if(*modeDiff)m_pFrm->m_pSocket->sendFmtData("MODE %s %s",caption(),modeDiff);
	m_pInput->setFocus();
}

void KviChannel::usersLabelRightClicked()
{
	if(!m_pMasksEditPopup)m_pMasksEditPopup = new QPopupMenu(this);
	m_pMasksEditPopup->clear();
	m_pMasksEditPopup->insertItem(__tr("Edit bans"),this,SLOT(doEditBans()));
	m_pMasksEditPopup->insertItem(__tr("Edit ban exceptions"),this,SLOT(doEditBanExceptions()));
	m_pMasksEditPopup->insertItem(__tr("Edit invite exceptions"),this,SLOT(doEditInviteExceptions()));
	m_pMasksEditPopup->popup(QCursor::pos());
}

void KviChannel::prepareToEditMasks()
{
	if(!m_pMaskEditor){
		m_pMaskEditor  = new KviMaskEditor(this);
		m_pMaskEditor->connectMe(this,SLOT(maskEditCompleted(const char *)));
	}
	QPoint p=mapFromGlobal(m_pUsersLabel->mapToGlobal(QPoint(0,m_pUsersLabel->height())));
	int hght = m_pMaskEditor->sizeHint().height();
	if(hght > (height() - m_pUsersLabel->height()))hght = height() - m_pUsersLabel->height();
	m_pMaskEditor->setGeometry(p.x(),p.y(),m_pUsersLabel->width(),hght);
	QTimer::singleShot(10,this,SLOT(absurdWorkaroundForMaskEditorFocus()));
}

void KviChannel::absurdWorkaroundForMaskEditorFocus()
{
	m_pMaskEditor->setFocus();
//	debug("focus widget = %s",g_pApp->focusWidget()->className());
}

void KviChannel::doEditBans()
{
	prepareToEditMasks();
	m_iLastMasksEdited = KVI_MASKEDIT_BANS;
	m_pMaskEditor->doEdit(m_pBanList,__tr("Active bans"),isMeOp());
}

void KviChannel::doEditBanExceptions()
{
	prepareToEditMasks();
	m_iLastMasksEdited = KVI_MASKEDIT_BANEXCEPTIONS;
	m_pMaskEditor->doEdit(m_pExceptionList,__tr("Active ban exceptions"),isMeOp());
}

void KviChannel::doEditInviteExceptions()
{
	prepareToEditMasks();
	m_iLastMasksEdited = KVI_MASKEDIT_INVITEEXCEPTIONS;
	m_pMaskEditor->doEdit(m_pInviteExceptionList,__tr("Active invite exceptions"),isMeOp());
}

void KviChannel::doEditMasks()
{
	switch(m_iLastMasksEdited){
		case KVI_MASKEDIT_BANS:             doEditBans();              break;
		case KVI_MASKEDIT_BANEXCEPTIONS:    doEditBanExceptions();     break;
		case KVI_MASKEDIT_INVITEEXCEPTIONS: doEditInviteExceptions();  break;
		default:                            doEditBans();              break;
	}
}

void KviChannel::maskEditCompleted(const char *masksToRemove)
{
	// Unban/unexc 3 masks per time...
	KviStr aFlag;
	switch(m_iLastMasksEdited){
		case KVI_MASKEDIT_BANS:             aFlag = "b";        break;
		case KVI_MASKEDIT_BANEXCEPTIONS:    aFlag = "e";        break;
		case KVI_MASKEDIT_INVITEEXCEPTIONS: aFlag = "I";        break;
		default:               /*What ?*/    return;            break;
	}
	if(*masksToRemove){
		KviStr tmp;
		KviStr masks;
		KviStr flags;
		int i=0;
		while(*masksToRemove){
			masksToRemove = kvi_extractToken(tmp,masksToRemove,' ');
			if(masks.hasData())masks+=" ";
			masks.append(tmp);
			flags.append(aFlag);
			i++;
			if(i == 3){
				m_pFrm->m_pSocket->sendFmtData("MODE %s -%s %s",caption(),flags.ptr(),masks.ptr());
				masks = "";
				flags = "";
				i = 0;
			}
		}
		if(i != 0)m_pFrm->m_pSocket->sendFmtData("MODE %s -%s %s",caption(),flags.ptr(),masks.ptr());
	}
}

void KviChannel::viewRightClicked(KviIrcView *)
{
	g_pChannelPopup->doPopup(this,SLOT(viewPopupClicked(const KviStr &)));
}

void KviChannel::viewPopupClicked(const KviStr &dataBuffer)
{
	m_pFrm->m_pUserParser->parseCommand(dataBuffer.ptr(),this);
}

void KviChannel::listBoxRightClicked()
{
	g_pUserlistPopup->doPopup(this,SLOT(userlistPopupClicked(const KviStr &)));
}

void KviChannel::userlistPopupClicked(const KviStr &dataBuffer)
{
//	debug("Userlist popup clicked : %s : channel %s",dataBuffer.ptr(),caption());
	m_pFrm->m_pUserParser->parseCommand(dataBuffer.ptr(),this);
}


//=============== applyOptions ================//

void KviChannel::applyLabelsOptions()
{
	if(!g_pOptions->m_pixLabelsBack->isNull()){
		m_pTopicLabel->setBackgroundPixmap(*(g_pOptions->m_pixLabelsBack));
		m_pModeLabel->setBackgroundPixmap(*(g_pOptions->m_pixLabelsBack));
		m_pUsersLabel->setBackgroundPixmap(*(g_pOptions->m_pixLabelsBack));
	} else {
		m_pTopicLabel->setBackgroundColor(g_pOptions->m_clrLabelsBack);
		m_pModeLabel->setBackgroundColor(g_pOptions->m_clrLabelsBack);
		m_pUsersLabel->setBackgroundColor(g_pOptions->m_clrLabelsBack);
	}
	QPalette pal(m_pTopicLabel->palette());
	pal.setColor(QPalette::Normal,QColorGroup::Text,g_pOptions->m_clrLabelsActiveFore);
	m_pUsersLabel->setPalette(pal);

	setTopicChangeEnabled(isMeOp() || (!isModeActive('t')));
	setModeChangeEnabled(isMeOp());

	m_pTopicLabel->setFont(g_pOptions->m_fntLabels);
	m_pModeLabel->setFont(g_pOptions->m_fntLabels);
	m_pUsersLabel->setFont(g_pOptions->m_fntLabels);
}

void KviChannel::setTopicChangeEnabled(bool bEnabled)
{
	m_bTopicChangeEnabled = bEnabled;
	QPalette pal(m_pTopicLabel->palette());
	pal.setColor(QPalette::Normal,QColorGroup::Text,
		bEnabled ? g_pOptions->m_clrLabelsActiveFore : g_pOptions->m_clrLabelsInactiveFore);
	m_pTopicLabel->setPalette(pal);
	m_pTopicLabel->update();
}

void KviChannel::setModeChangeEnabled(bool bEnabled)
{
	m_bModeChangeEnabled  = bEnabled;
	QPalette pal(m_pModeLabel->palette());
	pal.setColor(QPalette::Normal,QColorGroup::Text,
		bEnabled ? g_pOptions->m_clrLabelsActiveFore : g_pOptions->m_clrLabelsInactiveFore);
	m_pModeLabel->setPalette(pal);
	m_pModeLabel->update();
}

bool KviChannel::isMeOp()
{
	return m_pListBox->isOp(m_pFrm->m_global.szCurrentNick.ptr());
}

bool KviChannel::isOp(const char *nick)
{
	return m_pListBox->isOp(nick);
}

bool KviChannel::isVoice(const char *nick)
{
	return m_pListBox->isVoice(nick);
}

bool KviChannel::isModeActive(char modeFlag)
{
	QString txt=m_pModeLabel->text();
	return (txt.find(modeFlag) != -1);
}

void KviChannel::setModeFlag(char modeFlag,bool bSet)
{
	QString txt=m_pModeLabel->text();
	if(bSet){
		if(txt.find(modeFlag) == -1){
			txt+=modeFlag;
			m_pModeLabel->setText(txt);
		}
	} else {
		int idx = txt.find(modeFlag);
		if(idx != -1){
			txt.remove(idx,1);
			m_pModeLabel->setText(txt);
		}
	}
	if(modeFlag == 't')setTopicChangeEnabled(isMeOp() || (!isModeActive('t')));
}

void KviChannel::endOfNamesList()
{
	if(!(m_iChanSyncMask & KVI_CHANSYNC_HAVENAMES)){
		m_iChanSyncMask |= KVI_CHANSYNC_HAVENAMES;
		if(m_pListBox->count() == 1){
			// I'm alone!
			// Theoretically , no topic is set
			// And most servers will not send it
			m_iChanSyncMask |= KVI_CHANSYNC_HAVETOPIC;
		}
		checkSyncComplete();
	}
	repaintListBox();

}
void KviChannel::endOfWho()
{
	if(!(m_iChanSyncMask & KVI_CHANSYNC_HAVEWHO)){
		m_iChanSyncMask |= KVI_CHANSYNC_HAVEWHO;
		checkSyncComplete();
	}
}
void KviChannel::gotModeReply()
{
	if(!(m_iChanSyncMask & KVI_CHANSYNC_HAVEMODE)){
		m_iChanSyncMask |= KVI_CHANSYNC_HAVEMODE;
		checkSyncComplete();
	}
}
void KviChannel::gotTopicReply()
{
	if(!(m_iChanSyncMask & KVI_CHANSYNC_HAVETOPIC)){
		m_iChanSyncMask |= KVI_CHANSYNC_HAVETOPIC;
		checkSyncComplete();
	}
}
void KviChannel::gotBanlistReply()
{
	if(!(m_iChanSyncMask & KVI_CHANSYNC_HAVEBANS)){
		m_iChanSyncMask |= KVI_CHANSYNC_HAVEBANS;
		checkSyncComplete();
	}
}
void KviChannel::checkSyncComplete()
{
	bool bInSync = false;
	if(g_pOptions->m_bRequestWhoAfterEndOfNames){
		bInSync = (m_iChanSyncMask == (KVI_CHANSYNC_HAVEWHO | KVI_CHANSYNC_HAVEMODE | KVI_CHANSYNC_HAVETOPIC | KVI_CHANSYNC_HAVENAMES | KVI_CHANSYNC_HAVEBANS));
	} else {
		bInSync = (m_iChanSyncMask == (KVI_CHANSYNC_HAVEMODE | KVI_CHANSYNC_HAVETOPIC | KVI_CHANSYNC_HAVENAMES | KVI_CHANSYNC_HAVEBANS));
	}
	if(bInSync){
		int msecs = m_joinTime.msecsTo(QTime::currentTime());
		int secs  = msecs / 1000;
		msecs     = msecs % 1000;

		bool bHalt = false;
		if(g_pEventManager->eventEnabled(KviEvent_OnChannelSync)){
			KviStr tmp(KviStr::Format,"%s %d.%d",caption(),secs,msecs);
			bHalt = m_pFrm->m_pUserParser->callEvent(KviEvent_OnChannelSync,this,tmp);
		}

		if(g_pOptions->m_bShowChannelSyncTime && (!bHalt)){
			output(KVI_OUT_INTERNAL,__tr("Channel synchronized in %d.%d seconds"),secs,msecs);
		}
	}
}
void KviChannel::repaintListBox()
{
	m_pListBox->doRepaint();
	updateUsersLabel();
}

void KviChannel::join(const KviIrcUser &user,char bOp,char bVoice,bool bRepaint)
{
	m_pListBox->join(user,bOp,bVoice,bRepaint);
	if(bRepaint)updateUsersLabel();
}

void KviChannel::join(const char *nick,char bOp,char bVoice,bool bRepaint)
{
	KviIrcUser u(nick);
	m_pListBox->join(u,bOp,bVoice,bRepaint);
	if(bRepaint)updateUsersLabel();
}

void KviChannel::part(const char *nick)
{
	m_pListBox->part(nick);
	updateUsersLabel();
}

void KviChannel::part(const KviIrcUser &user)
{
	m_pListBox->part(user);
	updateUsersLabel();
}

bool KviChannel::nickChange(const KviIrcUser &nicker,const char *newNick)
{
	bool bRet = m_pListBox->nickChange(nicker,newNick);
	if(bRet)updateUsersLabel();
	return bRet;
}

bool KviChannel::op(const char *nick,bool bOp)
{
	bool bRet = m_pListBox->op(nick,bOp ? 1 : 0);
	updateUsersLabel();
	return bRet;
}

bool KviChannel::op(const KviIrcUser &user,bool bOp)
{
	bool bRet = m_pListBox->op(user,bOp ? 1 : 0);
	updateUsersLabel();
	return bRet;
}

bool KviChannel::voice(const char *nick,bool bVoice)
{
	bool bRet = m_pListBox->voice(nick,bVoice ? 1 : 0);
	updateUsersLabel();
	return bRet;
}

bool KviChannel::voice(const KviIrcUser &user,bool bVoice)
{
	bool bRet = m_pListBox->voice(user,bVoice ? 1 : 0);
	updateUsersLabel();
	return bRet;
}

bool KviChannel::isOnChannel(const KviIrcUser &user)
{
	return (m_pListBox->findUser(user) != 0);
}

bool KviChannel::isOnChannel(const char *nick)
{
	return (m_pListBox->findUser(nick) != 0);
}

void KviChannel::banMask(const char *banMask,const char *setBy,const char *setAt,bool bAdd)
{
	if(bAdd){
		// Make sure that it is not there yet...
		for(KviBanMask *s=m_pBanList->first();s;s=m_pBanList->next()){
			if(kvi_strEqualCI(s->mask.ptr(),banMask))return;
		}
		KviBanMask * m=new KviBanMask;
		m->mask = banMask;
		m->setBy = setBy;
		if(!setAt){
			m->setAt = QDateTime::currentDateTime().toString();
		} else m->setAt = setAt;
		m_pBanList->append(m);
		m_iBanCount++;
		updateUsersLabel();
	} else {
		for(KviBanMask *s=m_pBanList->first();s;s=m_pBanList->next()){
			if(kvi_strEqualCI(s->mask.ptr(),banMask)){
				m_pBanList->removeRef(s);
				m_iBanCount--;
				__range_valid(m_iBanCount >= 0);
				updateUsersLabel();
				return;
			}
		}
	}
}

void KviChannel::banExceptionMask(const char *banExcMask,const char *setBy,const char *setAt,bool bAdd)
{
	if(bAdd){
		// Make sure that it is not there yet...
		for(KviBanMask *s=m_pExceptionList->first();s;s=m_pExceptionList->next()){
			if(kvi_strEqualCI(s->mask.ptr(),banExcMask))return;
		}
		KviBanMask *m=new KviBanMask;
		m->mask = banExcMask;
		m->setBy = setBy;
		if(!setAt){
			m->setAt = QDateTime::currentDateTime().toString();
		} else m->setAt = setAt;
		m_pExceptionList->append(m);
		m_iExceptionCount++;
		updateUsersLabel();
	} else {
		for(KviBanMask *s=m_pExceptionList->first();s;s=m_pExceptionList->next()){
			if(kvi_strEqualCI(s->mask.ptr(),banExcMask)){
				m_pExceptionList->removeRef(s);
				m_iExceptionCount--;
				__range_valid(m_iExceptionCount >= 0);
				updateUsersLabel();
				return;
			}
		}
	}
}

void KviChannel::inviteExceptionMask(const char *inviteExcMask,const char *setBy,const char *setAt,bool bAdd)
{
	if(bAdd){
		// Make sure that it is not there yet...
		for(KviBanMask *s=m_pInviteExceptionList->first();s;s=m_pInviteExceptionList->next()){
			if(kvi_strEqualCI(s->mask.ptr(),inviteExcMask))return;
		}
		KviBanMask *m=new KviBanMask;
		m->mask = inviteExcMask;
		m->setBy = setBy;
		if(!setAt){
			m->setAt = QDateTime::currentDateTime().toString();
		} else m->setAt = setAt;
		m_pInviteExceptionList->append(m);
		m_iInviteExceptionCount++;
		updateUsersLabel();
	} else {
		for(KviBanMask *s=m_pInviteExceptionList->first();s;s=m_pInviteExceptionList->next()){
			if(kvi_strEqualCI(s->mask.ptr(),inviteExcMask)){
				m_pInviteExceptionList->removeRef(s);
				m_iInviteExceptionCount--;
				__range_valid(m_iInviteExceptionCount >= 0);
				updateUsersLabel();
				return;
			}
		}
	}
}
//#warning "apply other label options"

void KviChannel::setTopic(const char *text)
{
	QString t(text);
	QToolTip::remove(m_pTopicLabel);
	QToolTip::add(m_pTopicLabel,t);
	m_pTopicLabel->setText(t);
}

void KviChannel::appendTopic(KviStr &buffer)
{
	buffer.append(m_pTopicLabel->text());
}

void KviChannel::updateUsersLabel()
{
	KviStr str;
	if(m_iInviteExceptionCount){
		if(m_iExceptionCount){
			str.sprintf("%d [o:%d v:%d b:%d e:%d i:%d]",m_pListBox->count(),
				m_pListBox->opCount(),m_pListBox->voiceCount(),m_iBanCount,m_iExceptionCount,m_iInviteExceptionCount);
		} else {
			str.sprintf("%d [o:%d v:%d b:%d i:%d]",m_pListBox->count(),
				m_pListBox->opCount(),m_pListBox->voiceCount(),m_iBanCount,m_iInviteExceptionCount);
		}
	} else {
		if(m_iExceptionCount){
			str.sprintf("%d [o:%d v:%d b:%d e:%d]",m_pListBox->count(),
				m_pListBox->opCount(),m_pListBox->voiceCount(),m_iBanCount,m_iExceptionCount);
		} else {
			str.sprintf("%d [o:%d v:%d b:%d]",m_pListBox->count(),
				m_pListBox->opCount(),m_pListBox->voiceCount(),m_iBanCount);
		}
	}
	m_pUsersLabel->setText(QString(str.ptr()));
//	m_pUsersLabel->repaint(false);
}

void KviChannel::applyOptions()
{
	m_pView->setFont(g_pOptions->m_fntView);
	m_pView->setShowImages(g_pOptions->m_bShowImages,false);
	m_pView->setTimestamp(g_pOptions->m_bTimestamp);
	m_pView->setMaxBufferSize(g_pOptions->m_iViewMaxBufferSize);
	applyLabelsOptions();
}

//================ myIconPtr =================//

QPixmap * KviChannel::myIconPtr()
{
	return g_pWindowIcon[KVI_WND_ICON_CHANNEL];
}

//================ resizeEvent ===============//

void KviChannel::resizeEvent(QResizeEvent *)
{
	int inputSize = m_pInput->heightHint();
	int labelSize = m_pUsersLabel->sizeHint().height();
	m_pTopSplitter->setGeometry(0,0,width(),labelSize);
	m_pSplitter->setGeometry(0,labelSize,width(),height() - (inputSize+labelSize));
	m_pInput->setGeometry(0,height() - inputSize,width(),inputSize);
}

bool KviChannel::kickDumpChanStatus(const char *fName,const char *kicked,
		const char *kicker,const char *kicker_mask,const char *reason)
{
	QFile f(fName);
	if(f.open(IO_WriteOnly)){
		KviStr tmp(KviStr::Format,__tr("Kick log for %s on channel %s\n"),kicked,caption());
		f.writeBlock(tmp.ptr(),tmp.len());
		QDateTime tm(QDateTime::currentDateTime());
		KviStr d=tm.toString();
		tmp.sprintf(__tr("Written at %s\n\n"),d.ptr());
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp.sprintf(__tr("You were kicked from %s by %s [%s]\n"),caption(),kicker,kicker_mask);
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp.sprintf(__tr("Kick reason : %s\n\n"),reason);
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp = __tr("Channel topic:\n");
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp = m_pTopicLabel->text();
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp = __tr("\nChannel mode:\n");
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp = m_pModeLabel->text();
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp = __tr("\nChannel text buffer:\n");
		f.writeBlock(tmp.ptr(),tmp.len());
		m_pView->getTextBuffer(tmp);
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp.sprintf(__tr("You were kicked from %s by %s [%s] : %s\n\n"),caption(),kicker,kicker_mask,reason);
		f.writeBlock(tmp.ptr(),tmp.len());
		tmp = __tr("Channel user list:\n");
		f.writeBlock(tmp.ptr(),tmp.len());
		unsigned int num = 0;
		for(KviIrcUser *pU=m_pListBox->firstUser();pU;pU=m_pListBox->nextUser()){
			KviStr mask;
			pU->mask(mask);
			tmp.sprintf("%d : %s [%s]\n",num,pU->nick(),mask.ptr());
			f.writeBlock(tmp.ptr(),tmp.len());
			num++;
		}
		f.close();
		return true;	
	}
	return false;
}

#include "m_kvi_channel.moc"
