//
//   File : kvi_event.cpp (/usr/build/NEW_kvirc/kvirc/src/kvirc/kvi_event.cpp)
//   Last major modification : Tue Mar 30 1999 23:16:14 by Szymon Stefanek
//
//   This file is part of the KVirc irc client distribution
//   Copyright (C) 1999-2000 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#define _KVI_DEBUG_CHECK_RANGE_
#include "kvi_locale.h"
#include "kvi_debug.h"
#include "kvi_event.h"
#include "kvi_defines.h"

/* FIXME: "Protect events against self-modification" */

KviEventDescriptor KviEventManager::m_eventDescriptor[KVI_NUM_EVENTS]=
{
	{
		/*
			@event: OnStartup
			@short:
				New KVIrc server window opened
			@parameters:
				$1 = KVirc version string
			@window:
				Console
			@description:
				This event is 'fired' when a new KVirc server window is opened.<br>
				You can use it to setup your script , set internal variables or simply
				echo a 'script logo' to the console.<br>
				If you call <a href="halt.kvihelp">halt</a> in this event you will prevent KVirc from printing
				the version string on the console.
		*/
		"OnStartup","$1 version"
	},
	{
		/*
			@event: OnShutdown
			@short:
				KVIrc server window closed
			@parameters:
				None
			@window:
				Console
			@description:
				This event is 'fired' when a KVirc server window is closed.<br>
				Useful to save your script configuration.<br>
				Calling <a href="halt.kvihelp">halt</a> in this event has no particular meaning.
		*/
		"OnShutdown","none"
	},
	{
		/*
			@event: OnServerPing
			@short:
				Received a PING from server
			@parameters:
				$1 = server name<br>
				$2 = ping parameter
			@window:
				Console
			@description:
				Event executed when a PING from server has been received.<br>
				KVirc first replies to the ping (so you can not stop the reply),
				and then executes the event.<br>
				By calling <a href="halt.kvihelp">halt</a> in the event handler you will stop
				the PING notification output in the console (assuming that it has been enabled).
		*/
		"OnServerPing","$1 server,$2 parameter"
	},
	{
		/*
			@event: OnConnect
			@short:
				Connected to an IRC server
			@parameters:
				$1 = server hostname<br>
				$2 = server ip<br>
				$3 = port<br>
				$4 = network name
			@window:
				Console
			@description:
				This event is 'fired' when an irc connection has been estabilished.<br>
				No data has been sent to the server yet , so all the login operations
				are still to be done (even your nickname is still undefined!).<br>
				For this reason you should not attempt to send anything to the server
				unless you really know what you're doing.<br>
				Calling <a href="halp.kvihelp">halt</a> in this event will stop the 'Connected to ...' output in the console.
		*/
		"OnConnect","$1 server host,$2 server ip,$3 port,$4 network"
	},
	{
		/*
			@event: OnDisconnect
			@short:
				Disconnected from an IRC server
			@parameters:
				$1 = server hostname<br>
				$2 = server ip<br>
				$3 = port<br>
				$4 = network name
			@window:
				Console
			@description:
				This event is 'fired' when an irc connection has been terminated.<br>
				All the channel windows have been closed , and KVirc is in the middle of the "general system reset".<br>
				Obviously you should not attempt to send anything to the server : you are not connected.<br>
				You should use this event to reset your script state and stop any connection related asynchronous work
				that might be still running.<br>
				Calling <a href="halt.kvihelp">halt</a> in this event will avoid the 'Connection terminated ...' output in the console.
		*/
		"OnDisconnect","$1 server host,$2 server ip,$3 port,$4 network"
	},
	{
		/*
			@event: OnIrc
			@short:
				Login operations completed
			@parameters:
				$1 = nickname
			@window:
				Console
			@description:
				This event is 'fired' when all the login operations have been completed.<br>
				The nickname in $1 is the one that the server accepted (that may be also the secondary one or even auto-modified by KVirc).
				From this moment you are "really" on Irc.<br>
				Calling <a href="halt.kvihelp">halt</a> in this event will stop the 'Login operations complete : Happy ircing...' output in the console.<br>
				(In fact I wanted to call this event Event OnHappyIrcing).
		*/
		"OnIrc","$1 nickname"
	},
	{
		/*
			@event: OnJoin
			@short:
				An user joined a channel
			@parameters:
				$1 = channel name<br>
				$2 = joiner nickname<br>
				$3 = joiner username<br>
				$4 = joiner host
			@window:
				Joined channel
			@description:
				Event executed when an user (that is NOT you) joins a channel.<br>
				The special case in that YOU join a channel is handled by the event <a href="onMeJoin.kvihelp">OnMeJoin</a>.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you will avoid the '&lt;nick&gt; has joined &lt;channel&gt;...' output.
			@examples:
				Customize the join output:<br>
				<example>
					<a href="echo.kvihelp">echo</a> -i <a href="s_icon.kvihelp">$icon(join)</a> Signon : $2!$3@$4
					<a href="halt.kvihelp">halt</a>
				</example>
		*/
		"OnJoin","$1 channel name,$2 joiner nick,$3 joiner username,$4 joiner host"
	},
	{
		/*
			@event: OnChannelSync
			@short:
				Channel synchronisation completed
			@parameters:
				$1 = channel name<br>
				$2 = sync time (secs.msecs)
			@window:
				Channel
			@description:
				Event executed when a channel has been completly synchronized after a join.<br>
				When you join a channel KVirc has to request some information about it:<br>
				-Channel mode<br>
				-Topic<br>
				-List of users (and optionally the masks of all users)<br>
				-Ban list<br>
				This event fires when all of these informations have been received.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you will avoid the "Channel synchronized in ... secs" output.<br>
				<docsubtitle>Limitations</docsubtitle>
				The exception list state is still "undefined".<br>
				It is because we can not know if the server supports it.<br>
				So you can try to check it , but be aware that it may be not in sync.<br>
				<docsubtitle>More details</docsubtitle>
				When a RPL_ENDOFNAMES is received , and there is only one user on the channel
				the channel is assumed to be "new".<br>
				On "new" channels the topic is assumed to be "not set": KVirc will NOT wait for a RPL_TOPIC from server.<br>
				There is an exception for networks where a Channel Service is available
				and the channel is Registered.<br>
				In this case the topic may be unset in this event and set later by ChanServ.
		*/
		"OnChannelSync","$1 channel name,$2 synctime"
	},
	{
		/*
			@event: OnMeJoin
			@short:
				Joined a new channel
			@parameters:
				$1 = channel name
			@window:
				Joined channel
			@description:
				Event executed when YOU join a channel.<br>
				This is a special case for the event <a href="onJoin.kvihelp">OnJoin</a>.<br>
				Please note that this event is called just after the channel window has been created (in response to the
				server JOIN message).<br>
				If you try to check other users in the channel , you will strangely discover yourself ALONE!.<br>
				This is because the channel has not been synchronized yet.<br>
				For the same reason the topic and the channel mode is still undefined.<br>
				There is a special event that signals you channel synchronisation : <a href="onChannelSync.kvihelp">OnChannelSync</a>.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you will avoid the '&lt;nick&gt; has joined &lt;channel&gt;...' output.
		*/
		"OnMeJoin","$1 channel name"
	},
	{
		/*
			@event: OnPart
			@short:
				An user has left a channel
			@parameters:
				$1 = channel name<br>
				$2 = parter nickname<br>
				$3 = parter username<br>
				$4 = parter host<br>
				$5-= part message
			@window:
				Channel
			@description:
				Event executed when an user (that is NOT you) leaves a channel.<br>
				The special case in that YOU leave a channel is handled by the event <a href="onMePart.kvihelp">OnMePart</a>.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you will avoid the '&lt;nick&gt; has left &lt;channel&gt;...' output.
			@examples:
				Customize the part output:<br>
				<example>
					<a href="echo.kvihelp">echo</a> -i <a href="s_icon.kvihelp">$icon(part)</a> Signoff : $2!$3@$4 [$5-]
					<a href="halt.kvihelp">halt</a>
				</example>
		*/
		"OnPart","$1 channel name,$2 parter nick,$3 parter username,$4 parter host,$5- part message"
	},
	{
		/*
			@event: OnMePart
			@short:
				Left a channel
			@parameters:
				$1 = channel name<br>
				$2-= part message
			@window:
				Console
			@description:
				Event executed when YOU leave a channel.<br>
				This is a special case for the event <a href="onPart.kvihelp">OnPart</a>.<br>
				Called just after the channel window has been destroyed (in response to the
				server PART message).<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you will avoid the 'You have left &lt;channel&gt;...' output in the console.
		*/
		"OnMePart","$1 channel name,$2- part message"
	},
	{
		/*
			@event: OnKick
			@short:
				An user was kicked from a channel
			@parameters:
				$1 = channel name<br>
				$2 = kicked nick<br>
				$3 = kicker nick<br>
				$4 = kicker username<br>
				$5 = kicker host<br>
				$6-= kick reason
			@window:
				Channel
			@description:
				Event executed when an user (that is NOT you) has been kicked from a channel.<br>
				The special case in that YOU have been kicked is handled by the event <a href="onMeKick.kvihelp">OnMeKick</a>.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you will avoid the '&lt;nick&gt; has been kicked from &lt;channel&gt;...' output.
		*/
		"OnKick","$1 channel name,$2 kicked nick,$3 kicker nick,$4 kicker username,$5 kicker host,$6- kick reason"
	},
	{
		/*
			@event: OnMeKick
			@short:
				Kicked from a channel
			@parameters:
				$1 = channel name<br>
				$2 = kicker nick<br>
				$3 = kicker username<br>
				$4 = kicker host<br>
				$5-= kick reason
			@window:
				Console
			@description:
				Event executed when YOU have been kicked from a channel.<br>
				This is a special case for the event <a href="onKick.kvihelp">OnKick</a>.<br>
				Called just after the channel window has been closed (in response to the
				server KICK message).<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you will avoid the 'You have been kicked from &lt;channel&gt; by...' output in the console.
		*/
		"OnMeKick","$1 channel name,$2 kicker nick,$3 kicker username,$4 kicker host,$5- kick reason"
	},
	{
		/*
			@event: OnTopic
			@short:
				The topic of a channel was changed
			@parameters:
				$1 = channel name<br>
				$2 = source nickname<br>
				$3 = source mask<br>
				$4-= new topic
			@window:
				Channel
			@description:
				Event executed when an user (or server) sets a new topic for a channel.<br>
				By calling <a href="halp.kvihelp">halt</a> in this event you will avoid the '&lt;nick&gt; changes topic to ..' output.
		*/
		"OnTopic","$1 channel name,$2 source nick,$3 source mask,$4- new topic"
	},
	{
		/*
			@event: OnInvite
			@short:
				Received an INVITE message
			@parameters:
				$1 = inviter nick<br>
				$2 = inviter mask<br>
				$3 = channel<br>
				$4 = autojoin_is_enabled ?
			@window:
				Console
			@description:
				Event executed when an user invites you to join a channel.<br>
				If you call <a href="halp.kvihelp">halt</a> in this event , you will stop the "user foo invites you to join channel bar" output
				and the autojoin-in-invite feature.<br>
				$4 tells you if the feature is enabled by the user in the options dialog , so you can implement it
				by yourself.
			@examples:
			<example>
				<a href="echo.kvihelp">echo</a> -i <a href="s_icon.kvihelp">$icon(invite)</a> User $1 invites you to join channel $3
				<a href="if.kvihelp">if</a>($4)<a href="join.kvihelp">join</a> $3
				else echo The autojoin feature is disabled
				<a href="halt.kvihelp">halt</a>
			</example>
		*/
		"OnInvite","$1 inviter nick,$2 inviter mask,$3 channel,$4 autojoin_enabled?"
	},
	{ /* KviEvent_OnCTCP : halt stops the "CTCP ... REQUEST ..." output and the reply, bound as output */
		"OnCTCP","$1 source nick,$2 source mask,$3 target,$4 ctcp type,$5 reply_disabled?,$6- data"
	},
	{ /* KviEvent_OnCTCPReply : halt stops the "CTCP REPLY ..." output, bound as output*/
		"OnCTCPReply","$1 source nick,$2 source mask,$3 target, $4 ctcp type,$5- data"
	},
	{ /* KviEvent_OnCTCPFlood : halt stops the "CTCP FLOOD ..." output, bound to ActiveWindow or console, depending on options */
		"OnCTCPFlood","$1 source nick,$2 source mask,$3 target, $4 ctcp type,$5- data"
	},
	{ /* KviEvent_OnAction : halt stops the "user ..." output, bound as output */
		"OnAction","$1 source nick,$2 source mask,$3 target,$4 open_query_on_privmsg,$5- data"
	},

	{ /* KviEvent_OnOp : halt stops the "user sets mode +o user" output and the statusbat text, bound to chanel or console, dep. on options */
		"OnOp","$1 channel name,$2 source nick,$3 source mask,$4 dest nick"
	},
	{ /* KviEvent_OnDeop : halt stops the "user sets mode -o user" output and the statusbat text, bound to chanel or console, dep. on options */
		"OnDeOp","$1 channel name,$2 source nick,$3 source mask,$4 dest nick"
	},
	{ /* KviEvent_OnMeOp : halt stops the "user sets mode +o user" output and the statusbat text, bound to chanel or console, dep. on options */
		"OnMeOp","$1 channel name,$2 source nick,$3 source mask"
	},
	{ /* KviEvent_OnMeDeop : halt stops the "user sets mode -o user" output and the statusbat text, bound to chanel or console, dep. on options */
		"OnMeDeop","$1 channel name,$2 source nick,$3 source mask"
	},


	{ /* KviEvent_OnVoice : halt stops the "user sets mode +v user" output and the statusbat text, bound to chanel or console, dep. on options */
		"OnVoice","$1 channel name,$2 source nick,$3 source mask,$4 dest nick"
	},
	{ /* KviEvent_OnDevoice : halt stops the "user sets mode -v user" output and the statusbat text, bound to chanel or console, dep. on options */
		"OnDevoice","$1 channel name,$2 source nick,$3 source mask,$4 dest nick"
	},
	{ /* KviEvent_OnMeVoice : halt stops the "user sets mode +v user" output and the statusbat text, bound to chanel or console, dep. on options */
		"OnMeVoice","$1 channel name,$2 source nick,$3 source mask"
	},
	{ /* KviEvent_OnMeDevoice : halt stops the "user sets mode -v user" output and the statusbat text, bound to chanel or console, dep. on options */
		"OnMeDevoice","$1 channel name,$2 source nick,$3 source mask"
	},


	{ /* KviEvent_OnBan : halt stops the "user sets mode +b mask" output, bound to chan or console, dep. on options */
		"OnBan","$1 channel name,$2 source nick,$3 source mask,$4 dest mask[,$5 set_by,$6 set_at]"
	},
	{ /* KviEvent_OnUnban : halt stops the "user sets mode -b mask" output, bound to chan or console, dep. on options */
		"OnUnban","$1 channel name,$2 source nick,$3 source mask,$4 dest mask"
	},
	{ /* KviEvent_OnMeBan : halt stops the "user sets mode +b mask" output, bound to chan or console, dep. on options */
		"OnMeBan","$1 channel name,$2 source nick,$3 source mask,$4 dest mask[,$5 set_by,$6 set_at]"
	},
	{ /* KviEvent_OnMeUnban : halt stops the "user sets mode -b mask" output, bound to chan or console, dep. on options */
		"OnMeUnban","$1 channel name,$2 source nick,$3 source mask,$4 dest mask"
	},

	{ /* KviEvent_OnBanException : halt stops the "user sets mode +e mask" output, bound to chan or console, dep. on options */
		"OnBanException","$1 channel name,$2 source nick,$3 source mask,$4 dest mask[,$5 set_by,$6 set_at]"
	},
	{ /* KviEvent_OnBanExceptionRemove : halt stops the "user sets mode -e mask" output, bound to chan or console, dep. on options */
		"OnBanExceptionRemove","$1 channel name,$2 source nick,$3 source mask,$4 dest mask"
	},
	{ /* KviEvent_OnMeBanException : halt stops the "user sets mode +e mask" output, bound to chan or console, dep. on options */
		"OnMeBanException","$1 channel name,$2 source nick,$3 source mask,$4 dest mask[,$5 set_by,$6 set_at]"
	},
	{ /* KviEvent_OnMeBanExceptionRemove : halt stops the "user sets mode -e mask" output, bound to chan or console, dep. on options */
		"OnMeBanExceptionRemove","$1 channel name,$2 source nick,$3 source mask,$4 dest mask"
	},

	{
		/*
			@event: OnChannelMode
			@short:
				The channel mode changed
			@parameters:
				$1 = channel name<br>
				$2 = source nick/host<br>
				$3 = source mask (user@host)<br>
				$4 = mode flag(s)<br>
				$5 = parameter
			@window:
				Channel
			@description:
				Executed when a channel mode is changed by an user o a server.<br>
				In that case $4 contains a single mode flag (like +s or -t) and $5
				contains the (optional) parameter.<br>
				The event is firead also when you join a new channel and the
				server sends you the current channel mode.<br>
				In that case $2 and $3 contain the same string (the server name),
				and $4 contains the entire mode string (like +smtn) or an empty string
				if the channel is modeless.<br>
				If you call <a href="halt.kvihelp">halt</a> in this event , you will
				stop the "nick sets mode..." output.<br>
		*/
		"OnChannelMode","$1 channel name,$2 source nick,$3 source mask,$4 mode,$5 params"
	},

	{ /* KviEvent_OnChannelMessage : halt stops the Message output, bound to chan */
		/*
			@event: OnChannelMessage
			@short:
				A channel private message was received
			@parameters:
				$1 = source nickname<br>
				$2 = source mask (user@host)<br>
				$3- = notice text
			@window:
				Channel
			@description:
				Event fired when a channel private message arrives.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you
				will stop the message output.<br>
		*/
		"OnChannelMessage","$1 source nick,$2 source mask,$3- message"
	},
	{
		/*
			@event: OnMePrivateMessage
			@short:
				A private message was received
			@parameters:
				$1 = source nickname<br>
				$2 = source mask (user@host)<br>
				$3- = notice text
			@window:
				Query if exists , otherwise Console
			@description:
				Event fired when a private message directed to the user arrives.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you
				will stop the message output.<br>
				If no query window exists , halt will alsto stop the query creation.<br>
		*/
		"OnMePrivateMessage","$1 source nick,$2 source mask,$3- message"
	},

	{ 
		/*
			@event: OnServerNotice
			@short:
				A server notice was received
			@parameters:
				$1 = source server/service<br>
				$2 = target (nickname/channel)<br>
				$3- = notice text
			@window:
				Console
			@description:
				Event fired when a server/service notice arrives.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you
				will stop the message output.<br>
				The target of the output is depending on <a href="s_option.kvihelp">$option()</a>.
		*/
		"OnServerNotice","$1 source,$2 target,$3- notice"
	},
	{
		/*
			@event: OnChannelNotice
			@short:
				A channel notice was received
			@parameters:
				$1 = source nickname<br>
				$2 = source mask (user@host)<br>
				$3- = notice text
			@window:
				Channel
			@description:
				Event fired when a channel notice arrives.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you
				will stop the message output.<br>
		*/
		"OnChannelNotice","$1 source nick,$2 source mask,$3- notice text"
	},
	{
		/*
			@event: OnMeNotice
			@short:
				A private notice was received
			@parameters:
				$1 = source nickname<br>
				$2 = source mask (user@host)<br>
				$3- = notice text
			@window:
				Query if exists , otherwise Console
			@description:
				Event fired when a private notice arrives.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you
				will stop the message output.<br>
				If no query window exists , halt will stop also the query creation.<br>
		*/
		"OnMeNotice","$1 source nick,$2 source mask,$3- notice text"
	},

	{ /* KviEvent_OnNotify : halt stops the "user is on irc" output, bound to Console */
		"OnNotify","$1 nick, $2 isonline?"
	},

	{ /* KviEvent_OnInviteException : halt stops the "user sets mode +I mask" output, bound to chan or console, dep. on options */
		"OnInviteException","$1 channel name,$2 source nick,$3 source mask,$4 dest mask[,$5 set_by,$6 set_at]"
	},
	{ /* KviEvent_OnBanExceptionRemove : halt stops the "user sets mode -I mask" output, bound to chan or console, dep. on options */
		"OnInviteExceptionRemove","$1 channel name,$2 source nick,$3 source mask,$4 dest mask"
	},
	{ /* KviEvent_OnMeInviteException : halt stops the "user sets mode +I mask" output, bound to chan or console, dep. on options */
		"OnMeInviteException","$1 channel name,$2 source nick,$3 source mask,$4 dest mask[,$5 set_by,$6 set_at]"
	},
	{ /* KviEvent_OnMeInviteExceptionRemove : halt stops the "user sets mode -I mask" output, bound to chan or console, dep. on options */
		"OnMeInviteExceptionRemove","$1 channel name,$2 source nick,$3 source mask,$4 dest mask"
	},
	{ /* KviEvent_OnConsoleInput (46) : halt stops the "RAW to server..." output and the send, bound to console */
		"OnConsoleInput","$1 text"
	},
	{ /* KviEvent_OnChannelInput (47) : halt stops the <nick>... output and the send, bound to channel */
		"OnChannelInput","$1 window,$2 text"
	},
	{ /* KviEvent_OnQueryInput (48) : halt stops the <nick>... output and the send, bound to query */
		"OnQueryInput","$1 window,$2 text"
	},
	{ /* KviEvent_OnDCCInput (49) : halt stops the <nick>... output and the send, bound to dcc window */
		"OnDCCInput","$1 window,$2 text"
	},
	{
		/*
			@event: OnUnhandledLiteral
			@short:
				Message that KVirc ignores
			@parameters:
				$1 source mask<br>
				$2 literal command<br>
				$3- parameters
			@window:
				Console
			@description:
				Executed when KVIrc receives a literal message
				that is not understood.<br>
				This may happen for literal commands not handled yet by KVIrc.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you
				will stop the message output to console or active window
				(the target window is depending on an <a href="s_option.kvihelp">$option()</a>).<br>
		*/
		"OnUnhandledLiteral","$1 source mask, $2 literal command, $3- parameters"
	},
	{
		/*
			@event: OnUnhandledNumeric
			@short:
				Message that KVirc ignores
			@parameters:
				$1 source mask<br>
				$2 numeric reply<br>
				$3- parameters
			@window:
				Console
			@description:
				Executed when KVIrc receives a numeric message
				that is not understood.<br>
				This may happen for numeric replies not handled yet by KVIrc.<br>
				By calling <a href="halt.kvihelp">halt</a> in this event you
				will stop the message output to console or active window
				(the target window is depending on an <a href="s_option.kvihelp">$option()</a>).<br>
		*/
		"OnUnhandledNumeric","$1 source mask, $2 numeric reply, $3- parameters"
	},
	{
		/*
			@event: OnF1
			@short:
				F1 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
				A simple and funny example of its use could be a code like this:<br>
				<example>
					<a href="if.kvihelp">if</a>("<a href="s_commandline.kvihelp">$commandline</a>(<a href="s_window.kvihelp">$window</a>)" != "")<a href="window.kvihelp">window</a> <a href="s_window.kvihelp">$window</a> commandline <a href="s_b.kvihelp">$b</a><a href="s_u.kvihelp">$u</a><a href="s_commandline.kvihelp"><a href="s_upcase.kvihelp">$upcase</a>($commandline</a>(<a href="s_window.kvihelp">$window</a>))
				</example>
		*/
		"OnF1","none"
	},
	{
		/*
			@event: OnF2
			@short:
				F2 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF2","none"
	},
	{
		/*
			@event: OnF3
			@short:
				F3 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF3","none"
	},
	{
		/*
			@event: OnF4
			@short:
				F4 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF4","none"
	},
	{
		/*
			@event: OnF5
			@short:
				F5 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF5","none"
	},
	{
		/*
			@event: OnF6
			@short:
				F6 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF6","none"
	},
	{
		/*
			@event: OnF7
			@short:
				F7 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF7","none"
	},
	{
		/*
			@event: OnF8
			@short:
				F8 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF8","none"
	},
	{
		/*
			@event: OnF9
			@short:
				F9 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF9","none"
	},
	{
		/*
			@event: OnF10
			@short:
				F10 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF10","none"
	},
	{
		/*
			@event: OnF11
			@short:
				F11 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF11","none"
	},
	{
		/*
			@event: OnF12
			@short:
				F12 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnF12","none"
	},
	{
		/*
			@event: OnCtrl1
			@short:
				CTRL+1 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl1","none"
	},
	{
		/*
			@event: OnCtrl2
			@short:
				CTRL+2 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl2","none"
	},
	{
		/*
			@event: OnCtrl3
			@short:
				CTRL+3 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl3","none"
	},
	{
		/*
			@event: OnCtrl4
			@short:
				CTRL+4 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl4","none"
	},
	{
		/*
			@event: OnCtrl5
			@short:
				CTRL+5 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl5","none"
	},
	{
		/*
			@event: OnCtrl6
			@short:
				CTRL+6 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl6","none"
	},
	{
		/*
			@event: OnCtrl7
			@short:
				CTRL+7 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl7","none"
	},
	{
		/*
			@event: OnCtrl8
			@short:
				CTRL+8 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl8","none"
	},
	{
		/*
			@event: OnCtrl9
			@short:
				CTRL+9 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl9","none"
	},
	{
		/*
			@event: OnCtrl0
			@short:
				CTRL+0 shortcut key was pressed in a commandline input
			@parameters:
				none<br>
			@window:
				The one in which the key is pressed
			@description:
				Executed when a shortcut key is pressed in a commandline input.
		*/
		"OnCtrl0","none"
	},
	{
		/*
			@event: OnNickDoubleClicked
			@short:
				A nickname was doubleclicked in a channel (or console) listbox
			@parameters:
				$1 = nickname<br>
			@window:
				Channel or console in which the nick is clicked.<br>
			@description:
				Executed when a nickname in a listbox is double clicked.<br>
				By calling halt in this event you will stop the default action
				that is to call /QUERY nickname.<br>
		*/
		"OnNickDoubleClicked","nickname"
	},
	{ /* KviEvent_OnUserWindowInput (75) : halt stops the output and the send, bound to window */
		/*
			@event: OnUserWindowInput
			@short:
				Something was entered in a user window
			@parameters:
				$1 = window<br>
				$2- = text<br>
			@window:
				The window in which the text was entered.<br>
			@description:
				Executed when text was entered in a user window.<br>
				By calling halt in this event you will stop the text output.<br>
		*/
		"OnUserWindowInput","$1 window, $2- text"
	},
	{ /* KviEvent_OnUserWindowFileDrop (76) bound to window */
		/*
			@event: OnUserWindowFileDrop
			@short:
				A file was dropped (dnd) in a user window.
			@parameters:
				$1 = window<br>
				$2 = filename<br>
			@window:
				The userwindow in which the file was dropped.<br>
			@description:
				Executed when a file is droppend in a user window (view zone).<br>
		*/
		"OnUserWindowFileDrop","$1 window, $2 filename"
	},
	{ /* KviEvent_OnUserWindowSelected (77) bound to window */
		/*
			@event: OnUserWindowSelected
			@short:
				A list item was selected in a user window listbox
			@parameters:
				$1 = window<br>
				$2 = double-clicked?<br>
				$3 = index<br>
				$4- = item name<br>
			@window:
				The userwindow in which the item is selected.<br>
			@description:
				Executed when the selection in the listbox changes ($2 = 0).<br>
				Or when an item in the listbox is double-clicked ($2 = 1).<br>
		*/
		"OnUserWindowSelected","$1 window, $2 double-clicked, $3 index, $4- item name"
	},
	{ /* KviEvent_OnWhoisReply (78) bound to console */
		/*
			@event: OnWhoisReply
			@short:
				Whois reply from server
			@parameters:
				$1 = nickname<br>
				$2 = username<br>
				$3 = host<br>
			@window:
				Console<br>
			@description:
				Executed when Kvirc recieves a whois reply from the server.<br>
				Halt will stop all whois-output.<br>
		*/
		"OnWhoisReply","$1 nick, $2 username, $3 host"
	},
	{
		/*
			@event: OnDccChatMessage
			@short:
				Dcc chat remote message
			@parameters:
				$1 = nickname<br>
				$2 = username or "*" if not known<br>
				$3 = hostname or "*" if not known<br>
				$4 = ip_address<br>
				$5- = text<br>
			@window:
				Dcc chat window<br>
			@description:
				Executed when an user connected via dcc chat
				writes a line of text.<br>
				By calling 'halt' in this event you will avoid the output
				to the dcc window.<br>
		*/
		"OnDccChatMessage","$1 nick, $2 username or '*', $3 hostname or '*', $4 ip_address, $5- text"
	},
	{
		/*
			@event: OnDccChatConnected
			@short:
				Dcc chat connection estabilished
			@parameters:
				$1 = nickname<br>
				$2 = username or "*" if not known<br>
				$3 = hostname or "*" if not known<br>
				$4 = ip_address<br>
			@window:
				Dcc chat window<br>
			@description:
				Executed when a dcc chat connection is estabilished
		*/
		"OnDccChatConnected","$1 nick, $2 username or '*', $3 hostname or '*', $4 ip_address"
	},
	{
		/*
			@event: OnDccChatTerminated
			@short:
				Dcc chat connection terminated
			@parameters:
				$1 = nickname<br>
				$2 = username or "*" if not known<br>
				$3 = hostname or "*" if not known<br>
				$4 = ip_address<br>	
				$5- = error description (if any)<br>
			@window:
				Dcc chat window<br>
			@description:
				Executed when a dcc chat connection is terminated
				either by a remote close , error or local close.<br>
				This event does NOT fire if the connection was not estabilished yet.<br>
				By calling halt in this event you will disable the
				output "Connection terminated (host:port)"<br>
		*/
		"OnDccChatTerminated","$1 nick, $2 username or '*', $3 hostname or '*', $4 ip_address, $5- error message"
	},
	{
		/*
			@event: OnDccGetTransferComplete
			@short:
				Ad dcc get transfer has been succesfully completed
			@parameters:
				$1 = remote nickname<br>
				$2 = remote username or "*" if not known<br>
				$3 = remote hostname or "*" if not known<br>
				$4 = remote ip_address<br>	
				$5 = filename with path<br>
				$6 = length of the file in bytes
			@window:
				Dcc send window<br>
			@description:
				Executed when a dcc get transfer has been SUCCESSFULLY completed.<br>
		*/
		"OnDccGetTransferComplete","$1 nick, $2 username or '*', $3 hostname or '*', $4 ip_address, $5 filename, $6 length in bytes"
	},
	{
		/*
			@event: OnDccSendTransferComplete
			@short:
				Ad dcc send transfer has been succesfully completed
			@parameters:
				$1 = remote nickname<br>
				$2 = remote username or "*" if not known<br>
				$3 = remote hostname or "*" if not known<br>
				$4 = remote ip_address<br>	
				$5 = filename with path<br>
				$6 = length of the file in bytes
			@window:
				Dcc send window<br>
			@description:
				Executed when a dcc send transfer has been SUCCESSFULLY completed.<br>
		*/
		"OnDccSendTransferComplete","$1 nick, $2 username or '*', $3 hostname or '*', $4 ip_address, $5 filename, $6 length in bytes"
	},
	{
		/*
			@event: OnCtcpMultimedia
			@short:
				An user sent a CTCP Multimedia
			@parameters:
				$1 = source nickname<br>
				$2 = source mask<br>
				$3 = target (you or a channel name)<br>
				$4 = 1 if a query window should be created<br>	
				$5 = filename
			@window:
				Query or channel<br>
			@description:
				Executed when a remote user sends a CTCP MULTIMEDIA (SOUND) to you or a common channel.<br>
				Halt stops the output and the automatic query creation.<br>
		*/
		"OnCtcpMultimedia","$1 source nick,$2 source mask,$3 target,$4 open_query_on_privmsg,$5 filename"
	}
};


/* FIXME: "Order the events alphabetically" */

KviEventManager::KviEventManager()
{
	for(int i=0;i<KVI_NUM_EVENTS;i++){
		m_bEventEnabled[i]             = false;
		m_szEventBuffer[i]             = "";
#ifdef COMPILE_PLUGIN_SUPPORT
		m_pEventHandlerList[i]         = 0;
#endif
	}
}

KviEventManager::~KviEventManager()
{
#ifdef COMPILE_PLUGIN_SUPPORT
	for(int i=0;i<KVI_NUM_EVENTS;i++){
		if(m_pEventHandlerList[i] != 0){
			debug("WARNING : There are %u registered event handlers for event %d",m_pEventHandlerList[i]->count(),i);
			delete m_pEventHandlerList[i];
		}
	}
#endif
}

void KviEventManager::copyEventHandlers(KviEventManager * m)
{
	for(int i=0;i<KVI_NUM_EVENTS;i++){
		m_bEventEnabled[i] = m->m_bEventEnabled[i];
		m_szEventBuffer[i] = m->m_szEventBuffer[i];
	}
}

void KviEventManager::disableEmptyHandlers()
{
	for(int i=0;i<KVI_NUM_EVENTS;i++){
		m_szEventBuffer[i].stripWhiteSpace();
		if(m_szEventBuffer[i].isEmpty())m_bEventEnabled[i] = false;
	}
}

int KviEventManager::indexOfEvent(const char * eventName)
{
	int len = strlen(eventName);
	for(int i = 0;i< KVI_NUM_EVENTS;i++)
	{
		if(kvi_strEqualCIN(m_eventDescriptor[i].eventName,
			eventName,len))
		{
			if(len == (int)strlen(m_eventDescriptor[i].eventName))return i;
		}
	}
	return -1;
}

bool KviEventManager::load(const char *filename,KviStr &error)
{
	KviStr buffer;
	if(!kvi_loadFile(filename,buffer)){
		error = _i18n_("Unable to load file");
		return false;
	}
	if(!kvi_strEqualCSN(KVI_MAGIC_STRING,buffer.ptr(),KVI_MAGIC_STRING_LENGTH)){
		error = _i18n_("Invalid magic");
		return false; //not a kvirc file
	}
	buffer.cutLeft(KVI_MAGIC_STRING_LENGTH);
	buffer.stripLeftWhiteSpace(); //be flexible...allow some spaces...for manual editing by experts :)
	if(buffer.isEmpty())return true; //done

	while(buffer.hasData()){
		//[EVENT <eventname> <ENABLED|DISABLED> bytes <numofbytes>]\n
		//...
		//...
		//...
		//\n[ENDEVENT]\n

		// Event label
		if(!kvi_strEqualCIN("[EVENT ",buffer.ptr(),7)){
			error = _i18n_("Unrecognized [EVENT] label.");
			return false;
		}

		buffer.cutLeft(7);
		buffer.stripLeftWhiteSpace(); //be flexible...allow some spaces...for manual editing by experts :)

		// Event name
		KviStr eventName;
		if(!buffer.getToken(eventName,' ')){
			error = _i18n_("Syntax error in [EVENT] label");
			return false;
		} //there must be at least another token

		// Find the event
		int index = indexOfEvent(eventName.ptr());
		KviStr * ptrToEventBuffer;
		bool dummy;
		bool *ptrToEventEnableFlag;

		if((index < 0)||(index >= KVI_NUM_EVENTS)){
			debug("Unrecognized event handler %s , ignoring.",eventName.ptr());
			ptrToEventBuffer = &eventName; //<--put it here and ignore...
			ptrToEventEnableFlag = &dummy;
		} else {
			ptrToEventBuffer = &(m_szEventBuffer[index]);
			ptrToEventEnableFlag = &(m_bEventEnabled[index]);
		}

		buffer.stripLeftWhiteSpace(); //be flexible...allow some spaces...for manual editing by experts :)

		if(kvi_strEqualCIN("ENABLED ",buffer.ptr(),8)){
			buffer.cutLeft(8);
			*ptrToEventEnableFlag = true;
		} else if(kvi_strEqualCIN("DISABLED ",buffer.ptr(),9)){
			buffer.cutLeft(9);
			*ptrToEventEnableFlag = false;
		} else {
			error = _i18n_("Syntax error in [EVENT] label (event state)");
			return false;
		}

		buffer.stripLeftWhiteSpace(); //be flexible...allow some spaces...for manual editing by experts :)

		if(!kvi_strEqualCIN("bytes ",buffer.ptr(),6)){
			error = _i18n_("Syntax error in [EVENT] label (event length)");
			return false;
		}
		buffer.cutLeft(6);

		buffer.stripLeftWhiteSpace(); //be flexible...allow some spaces...for manual editing by experts :)

		if(!isdigit(*(buffer.ptr()))){
			error = _i18n_("Syntax error in [ERROR] label (digit expected)");
			return false;
		}

		KviStr numBytes;
		if(!buffer.getToken(numBytes,']')){
			error = _i18n_("Syntax error in [ERROR] label (can not extract alias length)");
			return false;
		}

		bool bOk = false;
		int bytes = numBytes.toInt(&bOk);
		if(!bOk){
			error = _i18n_("Unrecogized token ");
			error +=numBytes;
			return false;
		} //syntax error again...someone messed with the file
		buffer.cutLeft(1); //cut the \n
		if(buffer.len() < bytes){
			error = _i18n_("Alias buffer smaller than declared length (");
			error += numBytes;
			error +=")";
			return false;
		}

		ptrToEventBuffer->setStr(buffer.ptr(),bytes);
		buffer.cutLeft(bytes);

		if(!kvi_strEqualCIN("\n[ENDEVENT]",buffer.ptr(),11)){
			error = _i18n_("Unrecognized [ENDEVENT] label");
			return false;
		}
		buffer.cutLeft(11);

		buffer.stripLeftWhiteSpace(); //be flexible...allow some spaces...for manual editing by experts :)
	}
	disableEmptyHandlers();
	return true;
}

bool KviEventManager::save(const char *filename)
{
	QFile f(filename);
	if(!f.open(IO_WriteOnly | IO_Truncate))return false;

	f.writeBlock(KVI_MAGIC_STRING,KVI_MAGIC_STRING_LENGTH);

	for(int i = 0;i< KVI_NUM_EVENTS;i++){
		f.writeBlock("[EVENT ",7);
		f.writeBlock(m_eventDescriptor[i].eventName,strlen(m_eventDescriptor[i].eventName));
		if(m_bEventEnabled[i])f.writeBlock(" ENABLED",8);
		else f.writeBlock(" DISABLED",9);
		f.writeBlock(" bytes ",7);
		KviStr szNBytes(KviStr::Format,"%d",m_szEventBuffer[i].len());
		f.writeBlock(szNBytes.ptr(),szNBytes.len());
		f.writeBlock("]\n",2);
		f.writeBlock(m_szEventBuffer[i].ptr(),m_szEventBuffer[i].len());
		f.writeBlock("\n[ENDEVENT]\n",12);
	}

	f.close();
	return true;
}
